from __future__ import annotations

import busio
import microcontroller

class SPIDevice:
    """SPI Device Manager"""

    def __init__(
        self,
        spi: busio.SPI,
        chip_select: microcontroller.Pin,
        *,
        baudrate: int = 100000,
        polarity: int = 0,
        phase: int = 0,
        extra_clocks: int = 0,
    ) -> None:

        """
        Represents a single SPI device and manages locking the bus and the device address.

        :param ~busio.SPI spi: The SPI bus the device is on
        :param ~digitalio.DigitalInOut chip_select: The chip select pin object that implements the DigitalInOut API.
        :param bool cs_active_value: Set to true if your device requires CS to be active high. Defaults to false.
        :param int extra_clocks: The minimum number of clock cycles to cycle the bus after CS is high. (Used for SD cards.)

        Example::

            import busio
            import digitalio
            from board import *
            from adafruit_bus_device.spi_device import SPIDevice
            with busio.SPI(SCK, MOSI, MISO) as spi_bus:
                cs = digitalio.DigitalInOut(D10)
                device = SPIDevice(spi_bus, cs)
                bytes_read = bytearray(4)
                # The object assigned to spi in the with statements below
                # is the original spi_bus object. We are using the busio.SPI
                # operations busio.SPI.readinto() and busio.SPI.write().
                with device as spi:
                    spi.readinto(bytes_read)
                # A second transaction
                with device as spi:
                    spi.write(bytes_read)"""
    ...

    def __enter__(self) -> busio.SPI:
        """Starts a SPI transaction by configuring the SPI and asserting chip select."""
        ...
    def __exit__(self) -> None:
        """Ends a SPI transaction by deasserting chip select. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...



"""Basic digital pin support

The `digitalio` module contains classes to provide access to basic digital IO.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import digitalio
  import board

  pin = digitalio.DigitalInOut(board.LED)
  print(pin.value)

This example will initialize the the device, read
:py:data:`~digitalio.DigitalInOut.value` and then
:py:meth:`~digitalio.DigitalInOut.deinit` the hardware.

Here is blinky::

  import time
  import digitalio
  import board

  led = digitalio.DigitalInOut(board.LED)
  led.direction = digitalio.Direction.OUTPUT
  while True:
      led.value = True
      time.sleep(0.1)
      led.value = False
      time.sleep(0.1)

For the essentials of `digitalio`, see the `CircuitPython Essentials
Learn guide <https://learn.adafruit.com/circuitpython-essentials/circuitpython-digital-in-out>`_

For more information on using `digitalio`, see `this additional Learn guide
<https://learn.adafruit.com/circuitpython-digital-inputs-and-outputs>`_
"""

from __future__ import annotations

from typing import Optional

import microcontroller

class DriveMode:
    """Defines the drive mode of a digital pin"""

    def __init__(self) -> None:
        """Enum-like class to define the drive mode used when outputting
        digital values."""
        ...
    PUSH_PULL: DriveMode
    """Output both high and low digital values"""

    OPEN_DRAIN: DriveMode
    """Output low digital values but go into high z for digital high. This is
    useful for i2c and other protocols that share a digital line."""

class DigitalInOut:
    """Digital input and output

    A DigitalInOut is used to digitally control I/O pins. For analog control of
    a pin, see the :py:class:`analogio.AnalogIn` and
    :py:class:`analogio.AnalogOut` classes."""

    def __init__(self, pin: microcontroller.Pin) -> None:
        """Create a new DigitalInOut object associated with the pin. Defaults to input
        with no pull. Use :py:meth:`switch_to_input` and
        :py:meth:`switch_to_output` to change the direction.

        :param ~microcontroller.Pin pin: The pin to control"""
        ...
    def deinit(self) -> None:
        """Turn off the DigitalInOut and release the pin for other use."""
        ...
    def __enter__(self) -> DigitalInOut:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def switch_to_output(
        self, value: bool = False, drive_mode: DriveMode = DriveMode.PUSH_PULL
    ) -> None:
        """Set the drive mode and value and then switch to writing out digital
        values.

        :param bool value: default value to set upon switching
        :param ~digitalio.DriveMode drive_mode: drive mode for the output
        """
        ...
    def switch_to_input(self, pull: Optional[Pull] = None) -> None:
        """Set the pull and then switch to read in digital values.

        :param Pull pull: pull configuration for the input

        Example usage::

          import digitalio
          import board

          switch = digitalio.DigitalInOut(board.SLIDE_SWITCH)
          switch.switch_to_input(pull=digitalio.Pull.UP)
          # Or, after switch_to_input
          switch.pull = digitalio.Pull.UP
          print(switch.value)"""
        ...
    direction: Direction
    """The direction of the pin.

    Setting this will use the defaults from the corresponding
    :py:meth:`switch_to_input` or :py:meth:`switch_to_output` method. If
    you want to set pull, value or drive mode prior to switching, then use
    those methods instead."""

    value: bool
    """The digital logic level of the pin."""

    drive_mode: DriveMode
    """The pin drive mode. One of:

    - `digitalio.DriveMode.PUSH_PULL`
    - `digitalio.DriveMode.OPEN_DRAIN`"""

    pull: Optional[Pull]
    """The pin pull direction. One of:

    - `digitalio.Pull.UP`
    - `digitalio.Pull.DOWN`
    - `None`

    :raises AttributeError: if `direction` is :py:data:`~digitalio.Direction.OUTPUT`."""

class Direction:
    """Defines the direction of a digital pin"""

    def __init__(self) -> None:
        """Enum-like class to define which direction the digital values are
        going."""
        ...
    INPUT: Direction
    """Read digital data in"""

    OUTPUT: Direction
    """Write digital data out"""

class Pull:
    """Defines the pull of a digital input pin"""

    def __init__(self) -> None:
        """Enum-like class to define the pull value, if any, used while reading
        digital values in."""
        ...
    UP: Pull
    """When the input line isn't being driven the pull up can pull the state
    of the line high so it reads as true."""

    DOWN: Pull
    """When the input line isn't being driven the pull down can pull the
    state of the line low so it reads as false."""



"""Core font related data structures

.. note:: This module is intended only for low-level usage.  For working with
    fonts in CircuitPython see the `adafruit_bitmap_font library
    <https://github.com/adafruit/Adafruit_CircuitPython_Bitmap_Font>`_.
    For information on creating custom fonts for use in CircuitPython, see
    `this Learn guide <https://learn.adafruit.com/custom-fonts-for-pyportal-circuitpython-display>`_

"""

from __future__ import annotations

from typing import Optional, Tuple, Union

import displayio
from typing_extensions import Protocol  # for compat with python < 3.8

class FontProtocol(Protocol):
    """A protocol shared by `BuiltinFont` and classes in ``adafruit_bitmap_font``"""

    def get_bounding_box(self) -> Union[Tuple[int, int], Tuple[int, int, int, int]]:
        """Retrieve the maximum bounding box of any glyph in the font.

        The four element version is ``(width, height, x_offset, y_offset)``.
        The two element version is ``(width, height)``, in which
        ``x_offset`` and ``y_offset`` are assumed to be zero."""
        pass
    def get_glyph(self, codepoint: int) -> Optional[Glyph]:
        """Retrieve the Glyph for a given code point

        If the code point is not present in the font, `None` is returned."""
        pass

class BuiltinFont:
    """A font built into CircuitPython"""

    def __init__(self) -> None:
        """Creation not supported. Available fonts are defined when CircuitPython is built. See the
        `Adafruit_CircuitPython_Bitmap_Font <https://github.com/adafruit/Adafruit_CircuitPython_Bitmap_Font>`_
        library for dynamically loaded fonts."""
        ...
    bitmap: displayio.Bitmap
    """Bitmap containing all font glyphs starting with ASCII and followed by unicode. Use
    `get_glyph` in most cases. This is useful for use with `displayio.TileGrid` and
    `terminalio.Terminal`."""

    def get_bounding_box(self) -> Tuple[int, int]:
        """Returns the maximum bounds of all glyphs in the font in a tuple of two values: width, height."""
        ...
    def get_glyph(self, codepoint: int) -> Glyph:
        """Returns a `fontio.Glyph` for the given codepoint or None if no glyph is available."""
        ...

class Glyph:
    """Storage of glyph info"""

    def __init__(
        self,
        bitmap: displayio.Bitmap,
        tile_index: int,
        width: int,
        height: int,
        dx: int,
        dy: int,
        shift_x: int,
        shift_y: int,
    ) -> None:
        """Named tuple used to capture a single glyph and its attributes.

        :param bitmap: the bitmap including the glyph
        :param tile_index: the tile index within the bitmap
        :param width: the width of the glyph's bitmap
        :param height: the height of the glyph's bitmap
        :param dx: x adjustment to the bitmap's position
        :param dy: y adjustment to the bitmap's position
        :param shift_x: the x difference to the next glyph
        :param shift_y: the y difference to the next glyph"""
        ...



"""Lightweight 2D shapes for displays

The :py:attr:`vectorio` module provide simple filled drawing primitives for
use with `displayio`.

.. code-block:: python

    group = displayio.Group()

    palette = displayio.Palette(1)
    palette[0] = 0x125690

    circle = vectorio.Circle(pixel_shader=palette, radius=25, x=70, y=40)
    group.append(circle)

    rectangle = vectorio.Rectangle(pixel_shader=palette, width=40, height=30, x=55, y=45)
    group.append(rectangle)

    points=[(5, 5), (100, 20), (20, 20), (20, 100)]
    polygon = vectorio.Polygon(pixel_shader=palette, points=points, x=0, y=0)
    group.append(polygon)

"""

from __future__ import annotations

from typing import List, Tuple, Union

import displayio

class Circle:
    def __init__(
        self,
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        radius: int,
        x: int,
        y: int,
    ) -> None:
        """Circle is positioned on screen by its center point.

        :param Union[~displayio.ColorConverter,~displayio.Palette] pixel_shader: The pixel shader that produces colors from values
        :param int radius: The radius of the circle in pixels
        :param int x: Initial x position of the axis.
        :param int y: Initial y position of the axis.
        :param int color_index: Initial color_index to use when selecting color from the palette."""
    radius: int
    """The radius of the circle in pixels."""

    color_index: int
    """The color_index of the circle as 0 based index of the palette."""

    x: int
    """X position of the center point of the circle in the parent."""

    y: int
    """Y position of the center point of the circle in the parent."""

    location: Tuple[int, int]
    """(X,Y) position of the center point of the circle in the parent."""

    pixel_shader: Union[displayio.ColorConverter, displayio.Palette]
    """The pixel shader of the circle."""

class Polygon:
    def __init__(
        self,
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        points: List[Tuple[int, int]],
        x: int,
        y: int,
    ) -> None:
        """Represents a closed shape by ordered vertices. The path will be treated as
        'closed', the last point will connect to the first point.

        :param Union[~displayio.ColorConverter,~displayio.Palette] pixel_shader: The pixel
            shader that produces colors from values
        :param List[Tuple[int,int]] points: Vertices for the polygon
        :param int x: Initial screen x position of the 0,0 origin in the points list.
        :param int y: Initial screen y position of the 0,0 origin in the points list.
        :param int color_index: Initial color_index to use when selecting color from the palette."""
    points: List[Tuple[int, int]]
    """Vertices for the polygon."""

    color_index: int
    """The color_index of the polygon as 0 based index of the palette."""

    x: int
    """X position of the 0,0 origin in the points list."""

    y: int
    """Y position of the 0,0 origin in the points list."""

    location: Tuple[int, int]
    """(X,Y) position of the 0,0 origin in the points list."""

    pixel_shader: Union[displayio.ColorConverter, displayio.Palette]
    """The pixel shader of the polygon."""

class Rectangle:
    def __init__(
        self,
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        width: int,
        height: int,
        x: int,
        y: int,
    ) -> None:
        """Represents a rectangle by defining its bounds

        :param Union[~displayio.ColorConverter,~displayio.Palette] pixel_shader: The pixel shader that produces colors from values
        :param int width: The number of pixels wide
        :param int height: The number of pixels high
        :param int x: Initial x position of the top left corner.
        :param int y: Initial y position of the top left corner.
        :param int color_index: Initial color_index to use when selecting color from the palette."""
    width: int
    """The width of the rectangle in pixels."""

    height: int
    """The height of the rectangle in pixels."""

    color_index: int
    """The color_index of the rectangle in 1 based index of the palette."""

    x: int
    """X position of the top left corner of the rectangle in the parent."""

    y: int
    """Y position of the top left corner of the rectangle in the parent."""

    location: Tuple[int, int]
    """(X,Y) position of the top left corner of the rectangle in the parent."""

    pixel_shader: Union[displayio.ColorConverter, displayio.Palette]
    """The pixel shader of the rectangle."""



"""Hardware accelerated external bus access

The `busio` module contains classes to support a variety of serial
protocols.

When the microcontroller does not support the behavior in a hardware
accelerated fashion it may internally use a bitbang routine. However, if
hardware support is available on a subset of pins but not those provided,
then a RuntimeError will be raised. Use the `bitbangio` module to explicitly
bitbang a serial protocol on any general purpose pins.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import busio
  from board import *

  i2c = busio.I2C(SCL, SDA)
  print(i2c.scan())
  i2c.deinit()

This example will initialize the the device, run
:py:meth:`~busio.I2C.scan` and then :py:meth:`~busio.I2C.deinit` the
hardware. The last step is optional because CircuitPython automatically
resets hardware after a program finishes.

Note that drivers will typically handle communication if provided the bus
instance (such as ``busio.I2C(board.SCL, board.SDA)``), and that many of
the methods listed here are lower level functionalities that are needed
for working with custom drivers.

Tutorial for I2C and SPI:
https://learn.adafruit.com/circuitpython-basics-i2c-and-spi

Tutorial for UART:
https://learn.adafruit.com/circuitpython-essentials/circuitpython-uart-serial
"""

from __future__ import annotations

from typing import List, Optional

import microcontroller
from circuitpython_typing import ReadableBuffer, WriteableBuffer

class I2C:
    """Two wire serial protocol"""

    def __init__(
        self,
        scl: microcontroller.Pin,
        sda: microcontroller.Pin,
        *,
        frequency: int = 100000,
        timeout: int = 255,
    ) -> None:

        """I2C is a two-wire protocol for communicating between devices.  At the
        physical level it consists of 2 wires: SCL and SDA, the clock and data
        lines respectively.

        .. seealso:: Using this class directly requires careful lock management.
            Instead, use :class:`~adafruit_bus_device.I2CDevice` to
            manage locks.

        .. seealso:: Using this class to directly read registers requires manual
            bit unpacking. Instead, use an existing driver or make one with
            :ref:`Register <register-module-reference>` data descriptors.

        :param ~microcontroller.Pin scl: The clock pin
        :param ~microcontroller.Pin sda: The data pin
        :param int frequency: The clock frequency in Hertz
        :param int timeout: The maximum clock stretching timeut - (used only for
            :class:`bitbangio.I2C`; ignored for :class:`busio.I2C`)
        """
        ...
    def deinit(self) -> None:
        """Releases control of the underlying hardware so other classes can use it."""
        ...
    def __enter__(self) -> I2C:
        """No-op used in Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware on context exit. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def scan(self) -> List[int]:
        """Scan all I2C addresses between 0x08 and 0x77 inclusive and return a
        list of those that respond.

        :return: List of device ids on the I2C bus
        :rtype: list"""
        ...
    def try_lock(self) -> bool:
        """Attempts to grab the I2C lock. Returns True on success.

        :return: True when lock has been grabbed
        :rtype: bool"""
        ...
    def unlock(self) -> None:
        """Releases the I2C lock."""
        ...
    import sys
    def readfrom_into(
        self,
        address: int,
        buffer: WriteableBuffer,
        *,
        start: int = 0,
        end: int = sys.maxsize,
    ) -> None:
        """Read into ``buffer`` from the device selected by ``address``.
        At least one byte must be read.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed, but without copying the data.
        The number of bytes read will be the length of ``buffer[start:end]``.

        :param int address: 7-bit device address
        :param WriteableBuffer buffer: buffer to write into
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``"""
        ...
    import sys
    def writeto(
        self,
        address: int,
        buffer: ReadableBuffer,
        *,
        start: int = 0,
        end: int = sys.maxsize,
    ) -> None:
        """Write the bytes from ``buffer`` to the device selected by ``address`` and
        then transmit a stop bit.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``buffer[start:end]``.

        Writing a buffer or slice of length zero is permitted, as it can be used
        to poll for the existence of a device.

        :param int address: 7-bit device address
        :param ReadableBuffer buffer: buffer containing the bytes to write
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``
        """
        ...
    import sys
    def writeto_then_readfrom(
        self,
        address: int,
        out_buffer: ReadableBuffer,
        in_buffer: WriteableBuffer,
        *,
        out_start: int = 0,
        out_end: int = sys.maxsize,
        in_start: int = 0,
        in_end: int = sys.maxsize,
    ) -> None:
        """Write the bytes from ``out_buffer`` to the device selected by ``address``, generate no stop
        bit, generate a repeated start and read into ``in_buffer``. ``out_buffer`` and
        ``in_buffer`` can be the same buffer because they are used sequentially.

        If ``out_start`` or ``out_end`` is provided, then the buffer will be sliced
        as if ``out_buffer[out_start:out_end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``out_buffer[start:end]``.

        If ``in_start`` or ``in_end`` is provided, then the input buffer will be sliced
        as if ``in_buffer[in_start:in_end]`` were passed,
        The number of bytes read will be the length of ``out_buffer[in_start:in_end]``.
        :param int address: 7-bit device address
        :param ~circuitpython_typing.ReadableBuffer out_buffer: buffer containing the bytes to write
        :param ~circuitpython_typing.WriteableBuffer in_buffer: buffer to write into
        :param int out_start: beginning of ``out_buffer`` slice
        :param int out_end: end of ``out_buffer`` slice; if not specified, use ``len(out_buffer)``
        :param int in_start: beginning of ``in_buffer`` slice
        :param int in_end: end of ``in_buffer slice``; if not specified, use ``len(in_buffer)``
        """
        ...

class SPI:
    """A 3-4 wire serial protocol

    SPI is a serial protocol that has exclusive pins for data in and out of the
    main device.  It is typically faster than :py:class:`~bitbangio.I2C` because a
    separate pin is used to select a device rather than a transmitted
    address. This class only manages three of the four SPI lines: `!clock`,
    `!MOSI`, `!MISO`. Its up to the client to manage the appropriate
    select line, often abbreviated `!CS` or `!SS`. (This is common because
    multiple secondaries can share the `!clock`, `!MOSI` and `!MISO` lines
    and therefore the hardware.)"""

    def __init__(
        self,
        clock: microcontroller.Pin,
        MOSI: Optional[microcontroller.Pin] = None,
        MISO: Optional[microcontroller.Pin] = None,
        half_duplex: bool = False,
    ) -> None:

        """Construct an SPI object on the given pins.

        .. note:: The SPI peripherals allocated in order of desirability, if possible,
           such as highest speed and not shared use first. For instance, on the nRF52840,
           there is a single 32MHz SPI peripheral, and multiple 8MHz peripherals,
           some of which may also be used for I2C. The 32MHz SPI peripheral is returned
           first, then the exclusive 8MHz SPI peripheral, and finally the shared 8MHz
           peripherals.

        .. seealso:: Using this class directly requires careful lock management.
            Instead, use :class:`~adafruit_bus_device.SPIDevice` to
            manage locks.

        .. seealso:: Using this class to directly read registers requires manual
            bit unpacking. Instead, use an existing driver or make one with
            :ref:`Register <register-module-reference>` data descriptors.

        :param ~microcontroller.Pin clock: the pin to use for the clock.
        :param ~microcontroller.Pin MOSI: the Main Out Selected In pin.
        :param ~microcontroller.Pin MISO: the Main In Selected Out pin.
        :param bool half_duplex: True when MOSI is used for bidirectional data. False when SPI is full-duplex or simplex."""
        ...
    def deinit(self) -> None:
        """Turn off the SPI bus."""
        ...
    def __enter__(self) -> SPI:
        """No-op used by Context Managers.
        Provided by context manager helper."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def configure(
        self,
        *,
        baudrate: int = 100000,
        polarity: int = 0,
        phase: int = 0,
        bits: int = 8,
    ) -> None:
        """Configures the SPI bus. The SPI object must be locked.

        :param int baudrate: the desired clock rate in Hertz. The actual clock rate may be higher or lower
          due to the granularity of available clock settings.
          Check the `frequency` attribute for the actual clock rate.
        :param int polarity: the base state of the clock line (0 or 1)
        :param int phase: the edge of the clock that data is captured. First (0)
          or second (1). Rising or falling depends on clock polarity.
        :param int bits: the number of bits per word

        .. note:: On the SAMD21, it is possible to set the baudrate to 24 MHz, but that
           speed is not guaranteed to work. 12 MHz is the next available lower speed, and is
           within spec for the SAMD21.

        .. note:: On the nRF52840, these baudrates are available: 125kHz, 250kHz, 1MHz, 2MHz, 4MHz,
          and 8MHz.
          If you pick a a baudrate other than one of these, the nearest lower
          baudrate will be chosen, with a minimum of 125kHz.
          Two SPI objects may be created, except on the Circuit Playground Bluefruit,
          which allows only one (to allow for an additional I2C object)."""
        ...
    def try_lock(self) -> bool:
        """Attempts to grab the SPI lock. Returns True on success.

        :return: True when lock has been grabbed
        :rtype: bool"""
        ...
    def unlock(self) -> None:
        """Releases the SPI lock."""
        ...
    import sys
    def write(
        self, buffer: ReadableBuffer, *, start: int = 0, end: int = sys.maxsize
    ) -> None:
        """Write the data contained in ``buffer``. The SPI object must be locked.
        If the buffer is empty, nothing happens.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``buffer[start:end]``.

        :param ReadableBuffer buffer: write out bytes from this buffer
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, use ``len(buffer)``
        """
        ...
    import sys
    def readinto(
        self,
        buffer: WriteableBuffer,
        *,
        start: int = 0,
        end: int = sys.maxsize,
        write_value: int = 0,
    ) -> None:
        """Read into ``buffer`` while writing ``write_value`` for each byte read.
        The SPI object must be locked.
        If the number of bytes to read is 0, nothing happens.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]`` were passed.
        The number of bytes read will be the length of ``buffer[start:end]``.

        :param WriteableBuffer buffer: read bytes into this buffer
        :param int start: beginning of buffer slice
        :param int end: end of buffer slice; if not specified, it will be the equivalent value
            of ``len(buffer)`` and for any value provided it will take the value of
            ``min(end, len(buffer))``
        :param int write_value: value to write while reading
        """
        ...
    import sys
    def write_readinto(
        self,
        out_buffer: ReadableBuffer,
        in_buffer: WriteableBuffer,
        *,
        out_start: int = 0,
        out_end: int = sys.maxsize,
        in_start: int = 0,
        in_end: int = sys.maxsize,
    ) -> None:
        """Write out the data in ``out_buffer`` while simultaneously reading data into ``in_buffer``.
        The SPI object must be locked.

        If ``out_start`` or ``out_end`` is provided, then the buffer will be sliced
        as if ``out_buffer[out_start:out_end]`` were passed, but without copying the data.
        The number of bytes written will be the length of ``out_buffer[out_start:out_end]``.

        If ``in_start`` or ``in_end`` is provided, then the input buffer will be sliced
        as if ``in_buffer[in_start:in_end]`` were passed,
        The number of bytes read will be the length of ``out_buffer[in_start:in_end]``.

        The lengths of the slices defined by ``out_buffer[out_start:out_end]``
        and ``in_buffer[in_start:in_end]`` must be equal.
        If buffer slice lengths are both 0, nothing happens.

        :param ReadableBuffer out_buffer: write out bytes from this buffer
        :param WriteableBuffer in_buffer: read bytes into this buffer
        :param int out_start: beginning of ``out_buffer`` slice
        :param int out_end: end of ``out_buffer`` slice; if not specified, use ``len(out_buffer)``
        :param int in_start: beginning of ``in_buffer`` slice
        :param int in_end: end of ``in_buffer slice``; if not specified, use ``len(in_buffer)``
        """
        ...
    frequency: int
    """The actual SPI bus frequency. This may not match the frequency requested
    due to internal limitations."""

class UART:
    """A bidirectional serial protocol"""

    def __init__(
        self,
        tx: microcontroller.Pin,
        rx: microcontroller.Pin,
        *,
        baudrate: int = 9600,
        bits: int = 8,
        parity: Optional[Parity] = None,
        stop: int = 1,
        timeout: float = 1,
        receiver_buffer_size: int = 64,
    ) -> None:
        """A common bidirectional serial protocol that uses an an agreed upon speed
        rather than a shared clock line.

        :param ~microcontroller.Pin tx: the pin to transmit with, or ``None`` if this ``UART`` is receive-only.
        :param ~microcontroller.Pin rx: the pin to receive on, or ``None`` if this ``UART`` is transmit-only.
        :param ~microcontroller.Pin rts: the pin for rts, or ``None`` if rts not in use.
        :param ~microcontroller.Pin cts: the pin for cts, or ``None`` if cts not in use.
        :param ~microcontroller.Pin rs485_dir: the output pin for rs485 direction setting, or ``None`` if rs485 not in use.
        :param bool rs485_invert: rs485_dir pin active high when set. Active low otherwise.
        :param int baudrate: the transmit and receive speed.
        :param int bits:  the number of bits per byte, 5 to 9.
        :param Parity parity:  the parity used for error checking.
        :param int stop:  the number of stop bits, 1 or 2.
        :param float timeout:  the timeout in seconds to wait for the first character and between subsequent characters when reading. Raises ``ValueError`` if timeout >100 seconds.
        :param int receiver_buffer_size: the character length of the read buffer (0 to disable). (When a character is 9 bits the buffer will be 2 * receiver_buffer_size bytes.)

        *New in CircuitPython 4.0:* ``timeout`` has incompatibly changed units from milliseconds to seconds.
        The new upper limit on ``timeout`` is meant to catch mistaken use of milliseconds.

        .. note:: RS485 support on i.MX and Raspberry Pi RP2040 is implemented in software.
           The timing for the ``rs485_dir`` pin signal is done on a best-effort basis, and may not meet
           RS485 specifications intermittently.
        """
        ...
    def deinit(self) -> None:
        """Deinitialises the UART and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> UART:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def read(self, nbytes: Optional[int] = None) -> Optional[bytes]:
        """Read characters.  If ``nbytes`` is specified then read at most that many
        bytes. Otherwise, read everything that arrives until the connection
        times out. Providing the number of bytes expected is highly recommended
        because it will be faster.

        :return: Data read
        :rtype: bytes or None"""
        ...
    def readinto(self, buf: WriteableBuffer) -> Optional[int]:
        """Read bytes into the ``buf``. Read at most ``len(buf)`` bytes.

        :return: number of bytes read and stored into ``buf``
        :rtype: int or None (on a non-blocking error)

        *New in CircuitPython 4.0:* No length parameter is permitted."""
        ...
    def readline(self) -> bytes:
        """Read a line, ending in a newline character, or
           return None if a timeout occurs sooner, or
           return everything readable if no newline is found and timeout=0

        :return: the line read
        :rtype: bytes or None"""
        ...
    def write(self, buf: WriteableBuffer) -> Optional[int]:
        """Write the buffer of bytes to the bus.

        *New in CircuitPython 4.0:* ``buf`` must be bytes, not a string.

          :return: the number of bytes written
          :rtype: int or None"""
        ...
    baudrate: int
    """The current baudrate."""

    in_waiting: int
    """The number of bytes in the input buffer, available to be read"""

    timeout: float
    """The current timeout, in seconds (float)."""

    def reset_input_buffer(self) -> None:
        """Discard any unread characters in the input buffer."""
        ...

class Parity:
    """Enum-like class to define the parity used to verify correct data transfer."""

    ODD: int
    """Total number of ones should be odd."""

    EVEN: int
    """Total number of ones should be even."""



"""
The `ipaddress` module provides types for IP addresses. It is a subset of CPython's ipaddress
module.
"""

from __future__ import annotations

from typing import Union

def ip_address(obj: Union[int, str]) -> IPv4Address:
    """Return a corresponding IP address object or raise ValueError if not possible."""
    ...

class IPv4Address:
    """Encapsulates an IPv4 address."""

    def __init__(self, address: Union[int, str, bytes]) -> None:
        """Create a new IPv4Address object encapsulating the address value.

        The value itself can either be bytes or a string formatted address."""
        ...
    packed: bytes
    """The bytes that make up the address (read-only)."""

    version: int
    """4 for IPv4, 6 for IPv6"""

    def __eq__(self, other: object) -> bool:
        """Two Address objects are equal if their addresses and address types are equal."""
        ...
    def __hash__(self) -> int:
        """Returns a hash for the IPv4Address data."""
        ...



"""Support for audio samples"""

from __future__ import annotations

import typing
from typing import Optional

from circuitpython_typing import ReadableBuffer, WriteableBuffer

class RawSample:
    """A raw audio sample buffer in memory"""

    def __init__(
        self, buffer: ReadableBuffer, *, channel_count: int = 1, sample_rate: int = 8000
    ) -> None:
        """Create a RawSample based on the given buffer of signed values. If channel_count is more than
        1 then each channel's samples should alternate. In other words, for a two channel buffer, the
        first sample will be for channel 1, the second sample will be for channel two, the third for
        channel 1 and so on.

        :param ~circuitpython_typing.ReadableBuffer buffer: A buffer with samples
        :param int channel_count: The number of channels in the buffer
        :param int sample_rate: The desired playback sample rate

        Simple 8ksps 440 Hz sin wave::

          import audiocore
          import audioio
          import board
          import array
          import time
          import math

          # Generate one period of sine wav.
          length = 8000 // 440
          sine_wave = array.array("h", [0] * length)
          for i in range(length):
              sine_wave[i] = int(math.sin(math.pi * 2 * i / length) * (2 ** 15))

          dac = audioio.AudioOut(board.SPEAKER)
          sine_wave = audiocore.RawSample(sine_wave)
          dac.play(sine_wave, loop=True)
          time.sleep(1)
          dac.stop()"""
        ...
    def deinit(self) -> None:
        """Deinitialises the RawSample and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> RawSample:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    sample_rate: Optional[int]
    """32 bit value that dictates how quickly samples are played in Hertz (cycles per second).
    When the sample is looped, this can change the pitch output without changing the underlying
    sample. This will not change the sample rate of any active playback. Call ``play`` again to
    change it."""

class WaveFile:
    """Load a wave file for audio playback

    A .wav file prepped for audio playback. Only mono and stereo files are supported. Samples must
    be 8 bit unsigned or 16 bit signed. If a buffer is provided, it will be used instead of allocating
    an internal buffer, which can prevent memory fragmentation."""

    def __init__(self, file: typing.BinaryIO, buffer: WriteableBuffer) -> None:
        """Load a .wav file for playback with `audioio.AudioOut` or `audiobusio.I2SOut`.

        :param typing.BinaryIO file: Already opened wave file
        :param ~circuitpython_typing.WriteableBuffer buffer: Optional pre-allocated buffer,
          that will be split in half and used for double-buffering of the data.
          The buffer must be 8 to 1024 bytes long.
          If not provided, two 256 byte buffers are initially allocated internally.


        Playing a wave file from flash::

          import board
          import audiocore
          import audioio
          import digitalio

          # Required for CircuitPlayground Express
          speaker_enable = digitalio.DigitalInOut(board.SPEAKER_ENABLE)
          speaker_enable.switch_to_output(value=True)

          data = open("cplay-5.1-16bit-16khz.wav", "rb")
          wav = audiocore.WaveFile(data)
          a = audioio.AudioOut(board.A0)

          print("playing")
          a.play(wav)
          while a.playing:
            pass
          print("stopped")"""
        ...
    def deinit(self) -> None:
        """Deinitialises the WaveFile and releases all memory resources for reuse."""
        ...
    def __enter__(self) -> WaveFile:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    sample_rate: int
    """32 bit value that dictates how quickly samples are loaded into the DAC
    in Hertz (cycles per second). When the sample is looped, this can change
    the pitch output without changing the underlying sample."""

    bits_per_sample: int
    """Bits per sample. (read only)"""

    channel_count: int
    """Number of audio channels. (read only)"""



"""Support for edge counting

The `countio` module contains logic to read and count edge transistions

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info."""

from __future__ import annotations

from typing import Optional

import digitalio
import microcontroller

class Edge:
    """Enumerates which signal transitions can be counted."""

    def __init__(self) -> None:
        """Enum-like class to define which signal transitions to count."""
        ...
    RISE: Edge
    """Count the rising edges."""

    FALL: Edge
    """Count the falling edges."""

    RISE_AND_FALL: Edge
    """Count the rising and falling edges."""

class Counter:
    """Count the number of rising- and/or falling-edge transitions on a given pin."""

    def __init__(
        self,
        pin: microcontroller.Pin,
        *,
        edge: Edge = Edge.FALL,
        pull: Optional[digitalio.Pull] = None,
    ) -> None:
        """Create a Counter object associated with the given pin that counts
        rising- and/or falling-edge transitions. At least one of ``rise`` and ``fall`` must be True.
        The default is to count only falling edges, and is for historical backward compatibility.

        :param ~microcontroller.Pin pin: pin to monitor
        :param Edge edge: which edge transitions to count
        :param Optional[digitalio.Pull] pull: enable a pull-up or pull-down if not None


        For example::

            import board
            import countio

            # Count rising edges only.
            pin_counter = countio.Counter(board.D1, edge=Edge.RISE)
            # Reset the count after 100 counts.
            while True:
                if pin_counter.count >= 100:
                    pin_counter.reset()
                print(pin_counter.count)
        """
        ...
    def deinit(self) -> None:
        """Deinitializes the Counter and releases any hardware resources for reuse."""
    def __enter__(self) -> Counter:
        """No-op used by Context Managers."""
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
    count: int
    """The current count in terms of pulses."""

    def reset(self) -> None:
        """Resets the count back to 0."""



"""Access GIF-format images
"""

from __future__ import annotations

import typing
from typing import Union

import displayio
from circuitpython_typing import ReadableBuffer

class GifWriter:
    def __init__(
        self,
        file: Union[typing.BinaryIO, str],
        width: int,
        height: int,
        colorspace: displayio.Colorspace,
        loop: bool = True,
        dither: bool = False,
    ) -> None:
        """Construct a GifWriter object

        :param file: Either a file open in bytes mode, or the name of a file to open in bytes mode.
        :param width: The width of the image.  All frames must have the same width.
        :param height: The height of the image.  All frames must have the same height.
        :param colorspace: The colorspace of the image.  All frames must have the same colorspace.  The supported colorspaces are ``RGB565``, ``BGR565``, ``RGB565_SWAPPED``, ``BGR565_SWAPPED``, and ``L8`` (greyscale)
        :param loop: If True, the GIF is marked for looping playback
        :param dither: If True, and the image is in color, a simple ordered dither is applied.
        """
        ...
    def __enter__(self) -> GifWriter:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def deinit(self) -> None:
        """Close the underlying file."""
        ...
    def add_frame(self, bitmap: ReadableBuffer, delay: float = 0.1) -> None:
        """Add a frame to the GIF.

        :param bitmap: The frame data
        :param delay: The frame delay in seconds.  The GIF format rounds this to the nearest 1/100 second, and the largest permitted value is 655 seconds.
        """
        ...



"""Storage management

The `storage` provides storage management functionality such as mounting and
unmounting which is typically handled by the operating system hosting Python.
CircuitPython does not have an OS, so this module provides this functionality
directly.
For more information regarding using the `storage` module, refer to the `CircuitPython
Essentials Learn guide
<https://learn.adafruit.com/circuitpython-essentials/circuitpython-storage>`_.
"""

from __future__ import annotations

from typing import AnyStr, Iterator, Tuple, Union

def mount(filesystem: VfsFat, mount_path: str, *, readonly: bool = False) -> None:
    """Mounts the given filesystem object at the given path.

    This is the CircuitPython analog to the UNIX ``mount`` command.

    :param VfsFat filesystem: The filesystem to mount.
    :param str mount_path: Where to mount the filesystem.
    :param bool readonly: True when the filesystem should be readonly to CircuitPython.
    """
    ...

def umount(mount: Union[str, VfsFat]) -> None:
    """Unmounts the given filesystem object or if *mount* is a path, then unmount
    the filesystem mounted at that location.

    This is the CircuitPython analog to the UNIX ``umount`` command."""
    ...

def remount(
    mount_path: str,
    readonly: bool = False,
    *,
    disable_concurrent_write_protection: bool = False,
) -> None:
    """Remounts the given path with new parameters.

    :param str mount_path: The path to remount.
    :param bool readonly: True when the filesystem should be readonly to CircuitPython.
    :param bool disable_concurrent_write_protection: When True, the check that makes sure the
      underlying filesystem data is written by one computer is disabled. Disabling the protection
      allows CircuitPython and a host to write to the same filesystem with the risk that the
      filesystem will be corrupted."""
    ...

def getmount(mount_path: str) -> VfsFat:
    """Retrieves the mount object associated with the mount path"""
    ...

def erase_filesystem() -> None:
    """Erase and re-create the ``CIRCUITPY`` filesystem.

    On boards that present USB-visible ``CIRCUITPY`` drive (e.g., SAMD21 and SAMD51),
    then call `microcontroller.reset()` to restart CircuitPython and have the
    host computer remount CIRCUITPY.

    This function can be called from the REPL when ``CIRCUITPY``
    has become corrupted.

    .. warning:: All the data on ``CIRCUITPY`` will be lost, and
         CircuitPython will restart on certain boards."""
    ...

def disable_usb_drive() -> None:
    """Disable presenting ``CIRCUITPY`` as a USB mass storage device.
    By default, the device is enabled and ``CIRCUITPY`` is visible.
    Can be called in ``boot.py``, before USB is connected."""
    ...

def enable_usb_drive() -> None:
    """Enabled presenting ``CIRCUITPY`` as a USB mass storage device.
    By default, the device is enabled and ``CIRCUITPY`` is visible,
    so you do not normally need to call this function.
    Can be called in ``boot.py``, before USB is connected.

    If you enable too many devices at once, you will run out of USB endpoints.
    The number of available endpoints varies by microcontroller.
    CircuitPython will go into safe mode after running boot.py to inform you if
    not enough endpoints are available.
    """
    ...

class VfsFat:
    def __init__(self, block_device: str) -> None:
        """Create a new VfsFat filesystem around the given block device.

        :param block_device: Block device the the filesystem lives on"""
    label: str
    """The filesystem label, up to 11 case-insensitive bytes.  Note that
    this property can only be set when the device is writable by the
    microcontroller."""
    ...

    def mkfs(self) -> None:
        """Format the block device, deleting any data that may have been there"""
        ...
    def open(self, path: str, mode: str) -> None:
        """Like builtin ``open()``"""
        ...
    def ilistdir(
        self, path: str
    ) -> Iterator[Union[Tuple[AnyStr, int, int, int], Tuple[AnyStr, int, int]]]:
        """Return an iterator whose values describe files and folders within
        ``path``"""
        ...
    def mkdir(self, path: str) -> None:
        """Like `os.mkdir`"""
        ...
    def rmdir(self, path: str) -> None:
        """Like `os.rmdir`"""
        ...
    def stat(
        self, path: str
    ) -> Tuple[int, int, int, int, int, int, int, int, int, int]:
        """Like `os.stat`"""
        ...
    def statvfs(
        self, path: int
    ) -> Tuple[int, int, int, int, int, int, int, int, int, int]:
        """Like `os.statvfs`"""
        ...
    def mount(self, readonly: bool, mkfs: VfsFat) -> None:
        """Don't call this directly, call `storage.mount`."""
        ...
    def umount(self) -> None:
        """Don't call this directly, call `storage.umount`."""
        ...



"""Stack information and analysis"""

from __future__ import annotations

def max_stack_usage() -> int:
    """Return the maximum excursion of the stack so far."""
    ...

def stack_size() -> int:
    """Return the size of the entire stack.
    Same as in micropython.mem_info(), but returns a value instead
    of just printing it."""
    ...

def stack_usage() -> int:
    """Return how much stack is currently in use.
    Same as micropython.stack_use(); duplicated here for convenience."""
    ...



"""CAN bus access

The `canio` module contains low level classes to support the CAN bus
protocol.

CAN and Listener classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import canio
  from board import *

  can = canio.CAN(board.CAN_RX, board.CAN_TX, baudrate=1000000)
  message = canio.Message(id=0x0408, data=b"adafruit")
  can.send(message)
  can.deinit()

This example will write the data 'adafruit' onto the CAN bus to any
device listening for message id 0x0408.

A CAN bus involves a transceiver, which is often a separate chip with a "standby" pin.
If your board has a CAN_STANDBY pin, ensure to set it to an output with the value False
to enable the transceiver.

Other implementations of the CAN device may exist (for instance, attached
via an SPI bus).  If so their constructor arguments may differ, but
otherwise we encourage implementors to follow the API that the core uses.
"""

from __future__ import annotations

from types import TracebackType
from typing import Optional, Sequence, Type, Union

import microcontroller

class BusState:
    """The state of the CAN bus"""

    ERROR_ACTIVE: object
    """The bus is in the normal (active) state"""

    ERROR_WARNING: object
    """The bus is in the normal (active) state, but a moderate number of errors have occurred recently.

    .. note:: Not all implementations may use ``ERROR_WARNING``.  Do not rely on seeing ``ERROR_WARNING`` before ``ERROR_PASSIVE``."""

    ERROR_PASSIVE: object
    """The bus is in the passive state due to the number of errors that have occurred recently.

    This device will acknowledge packets it receives, but cannot transmit messages.
    If additional errors occur, this device may progress to BUS_OFF.
    If it successfully acknowledges other packets on the bus, it can return to ERROR_WARNING or ERROR_ACTIVE and transmit packets.
    """

    BUS_OFF: object
    """The bus has turned off due to the number of errors that have
    occurred recently.  It must be restarted before it will send or receive
    packets.  This device will neither send or acknowledge packets on the bus."""

class CAN:
    """CAN bus protocol"""

    def __init__(
        self,
        tx: microcontroller.Pin,
        rx: microcontroller.Pin,
        *,
        baudrate: int = 250000,
        loopback: bool = False,
        silent: bool = False,
        auto_restart: bool = False,
    ) -> None:
        """A common shared-bus protocol.  The rx and tx pins are generally
        connected to a transceiver which controls the H and L pins on a
        shared bus.

        :param ~microcontroller.Pin rx: the pin to receive with
        :param ~microcontroller.Pin tx: the pin to transmit with
        :param int baudrate: The bit rate of the bus in Hz.  All devices on the bus must agree on this value.
        :param bool loopback: When True  the ``rx`` pin's value is ignored, and the device receives the packets it sends.
        :param bool silent: When True the ``tx`` pin is always driven to the high logic level.  This mode can be used to "sniff" a CAN bus without interfering.
        :param bool auto_restart: If True, will restart communications after entering bus-off state
        """
        ...
    auto_restart: bool
    """If True, will restart communications after entering bus-off state"""

    baudrate: int
    """The baud rate (read-only)"""

    transmit_error_count: int
    """The number of transmit errors (read-only).  Increased for a detected transmission error, decreased for successful transmission.  Limited to the range from 0 to 255 inclusive.  Also called TEC."""

    receive_error_count: int
    """The number of receive errors (read-only).  Increased for a detected reception error, decreased for successful reception.  Limited to the range from 0 to 255 inclusive.  Also called REC."""

    state: BusState
    """The current state of the bus. (read-only)"""
    def restart(self) -> None:
        """If the device is in the bus off state, restart it."""
        ...
    def listen(
        self, matches: Optional[Sequence[Match]] = None, *, timeout: float = 10
    ) -> Listener:
        """Start receiving messages that match any one of the filters.

        Creating a listener is an expensive operation and can interfere with reception of messages by other listeners.

        There is an implementation-defined maximum number of listeners and limit to the complexity of the filters.

        If the hardware cannot support all the requested matches, a ValueError is raised.  Note that generally there are some number of hardware filters shared among all fifos.

        A message can be received by at most one Listener.  If more than one listener matches a message, it is undefined which one actually receives it.

        An empty filter list causes all messages to be accepted.

        Timeout dictates how long receive() and next() will block.

        Platform specific notes:

        SAM E5x supports two Listeners.  Filter blocks are shared between the two
        listeners.  There are 4 standard filter blocks and 4 extended filter blocks.
        Each block can either match 2 single addresses or a mask of addresses.
        The number of filter blocks can be increased, up to a hardware maximum, by
        rebuilding CircuitPython, but this decreases the CircuitPython free
        memory even if canio is not used.

        STM32F405 supports two Listeners.  Filter blocks are shared between the two listeners.
        There are 14 filter blocks.  Each block can match 2 standard addresses with
        mask or 1 extended address with mask.

        ESP32S2 supports one Listener.  There is a single filter block, which can either match a
        standard address with mask or an extended address with mask.
        """
        ...
    loopback: bool
    """True if the device was created in loopback mode, False
    otherwise (read-only)"""

    def send(self, message: Union[RemoteTransmissionRequest, Message]) -> None:
        """Send a message on the bus with the given data and id.
        If the message could not be sent due to a full fifo or a bus error condition, RuntimeError is raised.
        """
        ...
    silent: bool
    """True if the device was created in silent mode, False
    otherwise (read-only)"""

    def deinit(self) -> None:
        """Deinitialize this object, freeing its hardware resources"""
        ...
    def __enter__(self) -> CAN:
        """Returns self, to allow the object to be used in a `with` statement for resource control"""
        ...
    def __exit__(
        self,
        unused1: Optional[Type[BaseException]],
        unused2: Optional[BaseException],
        unused3: Optional[TracebackType],
    ) -> None:
        """Calls deinit()"""
        ...

class Listener:
    """Listens for CAN message

    `canio.Listener` is not constructed directly, but instead by calling
    `canio.CAN.listen`.

    In addition to using the `receive` method to retrieve a message or
    the `in_waiting` method to check for an available message, a
    listener can be used as an iterable, yielding messages until no
    message arrives within ``self.timeout`` seconds."""

    def receive(self) -> Optional[Union[RemoteTransmissionRequest, Message]]:
        """Reads a message, after waiting up to ``self.timeout`` seconds

        If no message is received in time, `None` is returned.  Otherwise,
        a `Message` or `RemoteTransmissionRequest` is returned."""
        ...
    def in_waiting(self) -> int:
        """Returns the number of messages (including remote
        transmission requests) waiting"""
        ...
    def __iter__(self) -> Listener:
        """Returns self

        This method exists so that `Listener` can be used as an
        iterable"""
        ...
    def __next__(self) -> Union[RemoteTransmissionRequest, Message]:
        """Reads a message, after waiting up to self.timeout seconds

        If no message is received in time, raises StopIteration.  Otherwise,
        a Message or  is returned.

        This method enables the `Listener` to be used as an
        iterable, for instance in a for-loop."""
        ...
    def deinit(self) -> None:
        """Deinitialize this object, freeing its hardware resources"""
        ...
    def __enter__(self) -> CAN:
        """Returns self, to allow the object to be used in a `with` statement for resource control"""
        ...
    def __exit__(
        self,
        unused1: Optional[Type[BaseException]],
        unused2: Optional[BaseException],
        unused3: Optional[TracebackType],
    ) -> None:
        """Calls deinit()"""
        ...
    timeout: float

class Match:
    """Describe CAN bus messages to match"""

    def __init__(
        self, id: int, *, mask: Optional[int] = None, extended: bool = False
    ) -> None:
        """Construct a Match with the given properties.

        If mask is not None, then the filter is for any id which matches all
        the nonzero bits in mask. Otherwise, it matches exactly the given id.
        If extended is true then only extended ids are matched, otherwise
        only standard ids are matched."""
    id: int
    """The id to match"""

    mask: int
    """The optional mask of ids to match"""

    extended: bool
    """True to match extended ids, False to match standard ides"""

class Message:
    def __init__(self, id: int, data: bytes, *, extended: bool = False) -> None:
        """Construct a Message to send on a CAN bus.

        :param int id: The numeric ID of the message
        :param bytes data: The content of the message
        :param bool extended: True if the message has an extended identifier, False if it has a standard identifier

        In CAN, messages can have a length from 0 to 8 bytes.
        """
        ...
    id: int
    """The numeric ID of the message"""

    data: bytes
    """The content of the message"""

    extended: bool
    """True if the message's id is an extended id"""

class RemoteTransmissionRequest:
    def __init__(self, id: int, length: int, *, extended: bool = False) -> None:
        """Construct a RemoteTransmissionRequest to send on a CAN bus.

        :param int id: The numeric ID of the requested message
        :param int length: The length of the requested message
        :param bool extended: True if the message has an extended identifier, False if it has a standard identifier

        In CAN, messages can have a length from 0 to 8 bytes.
        """
        ...
    id: int
    """The numeric ID of the message"""

    extended: bool
    """True if the message's id is an extended id"""

    length: int
    """The length of the requested message."""



"""USB CDC Serial streams

The `usb_cdc` module allows access to USB CDC (serial) communications.

On Windows, each `Serial` is visible as a separate COM port. The ports will often
be assigned consecutively, `console` first, but this is not always true.

On Linux, the ports are typically ``/dev/ttyACM0`` and ``/dev/ttyACM1``.
The `console` port will usually be first.

On MacOS, the ports are typically ``/dev/cu.usbmodem<something>``. The something
varies based on the USB bus and port used. The `console` port will usually be first.
"""

from __future__ import annotations

from typing import List, Optional

from circuitpython_typing import ReadableBuffer, WriteableBuffer

console: Optional[Serial]
"""The `console` `Serial` object is used for the REPL, and for `sys.stdin` and `sys.stdout`.
   `console` is ``None`` if disabled.

However, note that `sys.stdin` and `sys.stdout` are text-based streams,
and the `console` object is a binary stream.
You do not normally need to write to `console` unless you want to write binary data.
"""

data: Optional[Serial]
"""A `Serial` object that can be used to send and receive binary data to and from
the host.
Note that `data` is *disabled* by default. ``data`` is ``None`` if disabled."""

def disable() -> None:
    """Do not present any USB CDC device to the host.
    Can be called in ``boot.py``, before USB is connected.
    Equivalent to ``usb_cdc.enable(console=False, data=False)``."""
    ...

def enable(*, console: bool = True, data: bool = False) -> None:
    """Enable or disable each CDC device. Can be called in ``boot.py``, before USB is connected.

    :param console bool: Enable or disable the `console` USB serial device.
      True to enable; False to disable. Enabled by default.
    :param data bool: Enable or disable the `data` USB serial device.
      True to enable; False to disable. *Disabled* by default.

    If you enable too many devices at once, you will run out of USB endpoints.
    The number of available endpoints varies by microcontroller.
    CircuitPython will go into safe mode after running boot.py to inform you if
    not enough endpoints are available.
    """
    ...

class Serial:
    """Receives cdc commands over USB"""

    def __init__(self) -> None:
        """You cannot create an instance of `usb_cdc.Serial`.
        The available instances are in the ``usb_cdc.serials`` tuple."""
        ...
    def read(self, size: int = 1) -> bytes:
        """Read at most ``size`` bytes. If ``size`` exceeds the internal buffer size
        only the bytes in the buffer will be read. If `timeout` is > 0 or ``None``,
        and fewer than ``size`` bytes are available, keep waiting until the timeout
        expires or ``size`` bytes are available.

        :return: Data read
        :rtype: bytes"""
        ...
    def readinto(self, buf: WriteableBuffer) -> int:
        """Read bytes into the ``buf``.  If ``nbytes`` is specified then read at most
        that many bytes, subject to `timeout`.  Otherwise, read at most ``len(buf)`` bytes.

        :return: number of bytes read and stored into ``buf``
        :rtype: bytes"""
        ...
    def readline(self, size: int = -1) -> Optional[bytes]:
        r"""Read a line ending in a newline character ("\\n"), including the newline.
        Return everything readable if no newline is found and ``timeout`` is 0.
        Return ``None`` in case of error.

        This is a binary stream: the newline character "\\n" cannot be changed.
        If the host computer transmits "\\r" it will also be included as part of the line.

        :param int size: maximum number of characters to read. ``-1`` means as many as possible.
        :return: the line read
        :rtype: bytes or None"""
        ...
    def readlines(self) -> List[Optional[bytes]]:
        """Read multiple lines as a list, using `readline()`.

        .. warning:: If ``timeout`` is ``None``,
          `readlines()` will never return, because there is no way to indicate end of stream.

        :return: a list of the line read
        :rtype: list"""
        ...
    def write(self, buf: ReadableBuffer) -> int:
        """Write as many bytes as possible from the buffer of bytes.

        :return: the number of bytes written
        :rtype: int"""
        ...
    def flush(self) -> None:
        """Force out any unwritten bytes, waiting until they are written."""
        ...
    connected: bool
    """True if this Serial is connected to a host. (read-only)

    .. note:: The host is considered to be connected if it is asserting DTR (Data Terminal Ready).
      Most terminal programs and ``pyserial`` assert DTR when opening a serial connection.
      However, the C# ``SerialPort`` API does not. You must set ``SerialPort.DtrEnable``.
    """

    in_waiting: int
    """Returns the number of bytes waiting to be read on the USB serial input. (read-only)"""

    out_waiting: int
    """Returns the number of bytes waiting to be written on the USB serial output. (read-only)"""

    def reset_input_buffer(self) -> None:
        """Clears any unread bytes."""
        ...
    def reset_output_buffer(self) -> None:
        """Clears any unwritten bytes."""
        ...
    timeout: Optional[float]
    """The initial value of `timeout` is ``None``. If ``None``, wait indefinitely to satisfy
    the conditions of a read operation. If 0, do not wait. If > 0, wait only ``timeout`` seconds."""

    write_timeout: Optional[float]
    """The initial value of `write_timeout` is ``None``. If ``None``, wait indefinitely to finish
    writing all the bytes passed to ``write()``.If 0, do not wait.
    If > 0, wait only ``write_timeout`` seconds."""



"""Low-level routines for bitbanged LED matrices"""

from __future__ import annotations

from typing import Optional, Sequence

import digitalio
from circuitpython_typing import WriteableBuffer

class RGBMatrix:
    """Displays an in-memory framebuffer to a HUB75-style RGB LED matrix."""

    def __init__(
        self,
        *,
        width: int,
        bit_depth: int,
        rgb_pins: Sequence[digitalio.DigitalInOut],
        addr_pins: Sequence[digitalio.DigitalInOut],
        clock_pin: digitalio.DigitalInOut,
        latch_pin: digitalio.DigitalInOut,
        output_enable_pin: digitalio.DigitalInOut,
        doublebuffer: bool = True,
        framebuffer: Optional[WriteableBuffer] = None,
        height: int = 0,
        tile: int = 1,
        serpentine: bool = True,
    ) -> None:
        """Create a RGBMatrix object with the given attributes.  The height of
        the display is determined by the number of rgb and address pins and the number of tiles:
        ``len(rgb_pins) // 3 * 2 ** len(address_pins) * abs(tile)``.  With 6 RGB pins, 4
        address lines, and a single matrix, the display will be 32 pixels tall.  If the optional height
        parameter is specified and is not 0, it is checked against the calculated
        height.

        Up to 30 RGB pins and 8 address pins are supported.

        The RGB pins must be within a single "port" and performance and memory
        usage are best when they are all within "close by" bits of the port.
        The clock pin must also be on the same port as the RGB pins.  See the
        documentation of the underlying protomatter C library for more
        information.  Generally, Adafruit's interface boards are designed so
        that these requirements are met when matched with the intended
        microcontroller board.  For instance, the Feather M4 Express works
        together with the RGB Matrix Feather.

        The framebuffer is in "RGB565" format.

        "RGB565" means that it is organized as a series of 16-bit numbers
        where the highest 5 bits are interpreted as red, the next 6 as
        green, and the final 5 as blue.  The object can be any buffer, but
        `array.array` and ``ulab.ndarray`` objects are most often useful.
        To update the content, modify the framebuffer and call refresh.

        If a framebuffer is not passed in, one is allocated and initialized
        to all black.  In any case, the framebuffer can be retrieved
        by passing the RGBMatrix object to memoryview().

        If doublebuffer is False, some memory is saved, but the display may
        flicker during updates.

        A RGBMatrix is often used in conjunction with a
        `framebufferio.FramebufferDisplay`."""
    def deinit(self) -> None:
        """Free the resources (pins, timers, etc.) associated with this
        rgbmatrix instance.  After deinitialization, no further operations
        may be performed."""
        ...
    brightness: float
    """In the current implementation, 0.0 turns the display off entirely
    and any other value up to 1.0 turns the display on fully."""

    def refresh(self) -> None:
        """Transmits the color data in the buffer to the pixels so that
        they are shown."""
        ...
    width: int
    """The width of the display, in pixels"""

    height: int
    """The height of the display, in pixels"""



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT
import os
import site
from datetime import datetime
from typing import Dict, List

from setuptools import setup
from pathlib import Path

def local_scheme(version):
    return ""

STD_PACKAGES = set(('array', 'math', 'os', 'random', 'struct', 'sys', 'ssl', 'time'))

stub_root = Path(".")
stubs = [p.relative_to(stub_root) for p in stub_root.glob("*/*.pyi")]
packages = set(stub.parent.as_posix() for stub in stubs) - STD_PACKAGES
package_dir = dict((f"{package}-stubs", package)
    for package in packages)

def build_package_data() -> Dict[str, List[str]]:
    result = {}
    for package in packages:
        result[f"{package}-stubs"] = ["*.pyi", "*/*.pyi"]
    return result

package_data=build_package_data()
setup(
    name="circuitpython-stubs",
    description="PEP 561 type stubs for CircuitPython",
    url="https://github.com/adafruit/circuitpython",
    maintainer="CircuitPythonistas",
    maintainer_email="circuitpython@adafruit.com",
    author_email="circuitpython@adafruit.com",
    license="MIT",
    packages=list(package_data.keys()),
    package_data=package_data,
    package_dir = package_dir,
    setup_requires=["setuptools_scm", "setuptools>=38.6.0"],
    use_scm_version = {
        "root": "..",
        "relative_to": __file__,
        "local_scheme": local_scheme,
        "git_describe_command": "tools/describe --long",
    },
    zip_safe=False,
)



"""Audio output via digital PWM

The `audiopwmio` module contains classes to provide access to audio IO.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

Since CircuitPython 5, `Mixer`, `RawSample` and `WaveFile` are moved
to :mod:`audiocore`."""

from __future__ import annotations

from typing import Optional

import circuitpython_typing
import microcontroller

class PWMAudioOut:
    """Output an analog audio signal by varying the PWM duty cycle."""

    def __init__(
        self,
        left_channel: microcontroller.Pin,
        *,
        right_channel: Optional[microcontroller.Pin] = None,
        quiescent_value: int = 0x8000,
    ) -> None:
        """Create a PWMAudioOut object associated with the given pin(s). This allows you to
        play audio signals out on the given pin(s).  In contrast to mod:`audioio`,
        the pin(s) specified are digital pins, and are driven with a device-dependent PWM
        signal.

        :param ~microcontroller.Pin left_channel: The pin to output the left channel to
        :param ~microcontroller.Pin right_channel: The pin to output the right channel to
        :param int quiescent_value: The output value when no signal is present. Samples should start
            and end with this value to prevent audible popping.

        Simple 8ksps 440 Hz sin wave::

          import audiocore
          import audiopwmio
          import board
          import array
          import time
          import math

          # Generate one period of sine wav.
          length = 8000 // 440
          sine_wave = array.array("H", [0] * length)
          for i in range(length):
              sine_wave[i] = int(math.sin(math.pi * 2 * i / length) * (2 ** 15) + 2 ** 15)

          dac = audiopwmio.PWMAudioOut(board.SPEAKER)
          sine_wave = audiocore.RawSample(sine_wave, sample_rate=8000)
          dac.play(sine_wave, loop=True)
          time.sleep(1)
          dac.stop()

        Playing a wave file from flash::

          import board
          import audiocore
          import audiopwmio
          import digitalio

          # Required for CircuitPlayground Express
          speaker_enable = digitalio.DigitalInOut(board.SPEAKER_ENABLE)
          speaker_enable.switch_to_output(value=True)

          data = open("cplay-5.1-16bit-16khz.wav", "rb")
          wav = audiocore.WaveFile(data)
          a = audiopwmio.PWMAudioOut(board.SPEAKER)

          print("playing")
          a.play(wav)
          while a.playing:
            pass
          print("stopped")"""
        ...
    def deinit(self) -> None:
        """Deinitialises the PWMAudioOut and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> PWMAudioOut:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def play(
        self, sample: circuitpython_typing.AudioSample, *, loop: bool = False
    ) -> None:
        """Plays the sample once when loop=False and continuously when loop=True.
        Does not block. Use `playing` to block.

        Sample must be an `audiocore.WaveFile`, `audiocore.RawSample`, `audiomixer.Mixer` or `audiomp3.MP3Decoder`.

        The sample itself should consist of 16 bit samples. Microcontrollers with a lower output
        resolution will use the highest order bits to output."""
        ...
    def stop(self) -> None:
        """Stops playback and resets to the start of the sample."""
        ...
    playing: bool
    """True when an audio sample is being output even if `paused`. (read-only)"""

    def pause(self) -> None:
        """Stops playback temporarily while remembering the position. Use `resume` to resume playback."""
        ...
    def resume(self) -> None:
        """Resumes sample playback after :py:func:`pause`."""
        ...
    paused: bool
    """True when playback is paused. (read-only)"""



"""Native helpers for driving parallel displays"""

from __future__ import annotations

from typing import Optional

import microcontroller
from circuitpython_typing import ReadableBuffer

class ParallelBus:
    """Manage updating a display over 8-bit parallel bus in the background while Python code runs. This
    protocol may be refered to as 8080-I Series Parallel Interface in datasheets. It doesn't handle
    display initialization."""

    def __init__(
        self,
        *,
        data0: microcontroller.Pin,
        command: microcontroller.Pin,
        chip_select: microcontroller.Pin,
        write: microcontroller.Pin,
        read: Optional[microcontroller.Pin],
        reset: Optional[microcontroller.Pin] = None,
        frequency: int = 30_000_000,
    ) -> None:
        """Create a ParallelBus object associated with the given pins. The bus is inferred from data0
        by implying the next 7 additional pins on a given GPIO port.

        The parallel bus and pins are then in use by the display until `displayio.release_displays()`
        is called even after a reload. (It does this so CircuitPython can use the display after your
        code is done.) So, the first time you initialize a display bus in code.py you should call
        :py:func:`displayio.release_displays` first, otherwise it will error after the first code.py run.

        :param microcontroller.Pin data_pins: A list of data pins.  Specify exactly one of ``data_pins`` or ``data0``.
        :param microcontroller.Pin data0: The first data pin. The rest are implied
        :param microcontroller.Pin command: Data or command pin
        :param microcontroller.Pin chip_select: Chip select pin
        :param microcontroller.Pin write: Write pin
        :param microcontroller.Pin read: Read pin, optional
        :param microcontroller.Pin reset: Reset pin, optional
        :param int frequency: The communication frequency in Hz for the display on the bus"""
        ...
    def reset(self) -> None:
        """Performs a hardware reset via the reset pin. Raises an exception if called when no reset pin
        is available."""
        ...
    def send(self, command: int, data: ReadableBuffer) -> None:
        """Sends the given command value followed by the full set of data. Display state, such as
        vertical scroll, set via ``send`` may or may not be reset once the code is done."""
        ...



from __future__ import annotations

import ulab
import ulab.numpy
import ulab.scipy

"""Linear algebra functions"""

def solve_triangular(
    A: ulab.numpy.ndarray, b: ulab.numpy.ndarray, lower: bool
) -> ulab.numpy.ndarray:
    """
    :param ~ulab.numpy.ndarray A: a matrix
    :param ~ulab.numpy.ndarray b: a vector
    :param ~bool lower: if true, use only data contained in lower triangle of A, else use upper triangle of A
    :return: solution to the system A x = b. Shape of return matches b
    :raises TypeError: if A and b are not of type ndarray and are not dense
    :raises ValueError: if A is a singular matrix

    Solve the equation A x = b for x, assuming A is a triangular matrix"""
    ...

def cho_solve(L: ulab.numpy.ndarray, b: ulab.numpy.ndarray) -> ulab.numpy.ndarray:
    """
    :param ~ulab.numpy.ndarray L: the lower triangular, Cholesky factorization of A
    :param ~ulab.numpy.ndarray b: right-hand-side vector b
    :return: solution to the system A x = b. Shape of return matches b
    :raises TypeError: if L and b are not of type ndarray and are not dense

    Solve the linear equations A x = b, given the Cholesky factorization of A as input"""
    ...
